
#include "server.hpp"

#include <string>
#include <iostream>
#include <map>

#include <boost/asio.hpp>
#include <boost/thread.hpp>

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "tcpinfo.hpp"

extern std::map<int, boost::posix_time::ptime> start_times;
extern std::map<int, boost::posix_time::ptime> end_times;

namespace {

void receive_loop(int sock, int idx);
void signal_terminate(int signal);
void setup_signals();

}

void start_server(const std::string& address, unsigned short port)
{
    setup_signals();
    int sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    if (sock == -1)
    {
        std::cerr << "Error opening socket: " << std::strerror(errno) << std::endl;

        return;
    }

    sockaddr_in addr = sockaddr_in();
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr(address.c_str());
    addr.sin_port = htons(port);

    if (bind(sock, reinterpret_cast<sockaddr*>(&addr), sizeof(addr)) != 0)
    {
        std::cerr << "Error binding the socket: " << std::strerror(errno) << std::endl;
        close(sock);

        return;
    }

    if (listen(sock, 1024) != 0)
    {
        std::cerr << "Error creating the listening queue for the socket: " << std::strerror(errno) << std::endl;
        close(sock);

        return;
    }

    int idx = 0;

    while (true)
    {
        int s = accept(sock, 0, 0);

        if (s == -1)
        {
            std::cerr << "Connection error: " << std::strerror(errno) << std::endl;

            continue;
        }

        start_times.insert(std::make_pair(idx, boost::posix_time::microsec_clock::local_time()));
        boost::thread th(receive_loop, s, idx);
        th.detach();
	++idx;
    }
}

namespace {

void receive_loop(int sock, int idx) 
{
    char buf[1024] = { 0 };
    size_t tot_len = 0;

    while (true)
    {
        size_t len = recv(sock, buf, sizeof(buf), 0);

        if (len == -1)
        {
            std::cerr << "Error receiving packets: " << std::strerror(errno) << std::endl;

            break;
        }

        tot_len += len;

        //info.display(sock, tot_len);

        if (len == 0)
        {
            end_times.insert(std::make_pair(idx, boost::posix_time::microsec_clock::local_time()));

            break;
        }
    }

    close(sock);
}

void signal_terminate(int signal)
{
    typedef std::map<int, boost::posix_time::ptime>::const_iterator Iterator;
    double tot_times = 0;
    double num = 0;

    for (Iterator i = start_times.begin(); i!= start_times.end(); ++i)
    {
        Iterator e = end_times.find(i->first);

        if (e == end_times.end())
        {
            continue;
        }

        tot_times += (e->second - i->second).total_microseconds();
        ++num;
    }

    std::cout << "Average time: " << (tot_times / num) << " calculated on " << num << " samples" << std::endl;
}

void setup_signals()
{
    signal(SIGQUIT, signal_terminate);
	signal(SIGILL, signal_terminate);
	signal(SIGTRAP, signal_terminate);
	signal(SIGABRT, signal_terminate);
	signal(SIGIOT, signal_terminate);
	signal(SIGBUS, signal_terminate);
	signal(SIGFPE, signal_terminate);
	signal(SIGSEGV, signal_terminate);
	signal(SIGSTKFLT, signal_terminate);
	signal(SIGCONT, signal_terminate);
	signal(SIGPWR, signal_terminate);
	signal(SIGSYS, signal_terminate);

    /*sigtermSA.sa_handler = signal_terminate;
    sigtermSA.sa_flags   = 0;
    sigemptyset(&sigtermSA.sa_mask);
    sigaction( SIGTERM, &sigtermSA, NULL );

    sigusr1SA.sa_handler = signal_terminate;
    sigusr1SA.sa_flags   = 0;
    sigemptyset(&sigusr1SA.sa_mask);
    sigaction( SIGUSR1, &sigusr1SA, NULL );*/
}

} // Anonymous namespace
